from ._dataset_class import Era5, normalize, unnormalize
from ._chunkers import ContinuousChunker, SingleChunker, SingleRandomWrapChunker, Offset
from ._auxiliaries import Coordinate, CoordinateBundle

__all__ = [
    "Era5",
    "ContinuousChunker",
    "SingleChunker",
    "SingleRandomWrapChunker",
    "Offset",
    "Coordinate",
    "CoordinateBundle",
    "normalize",
    "unnormalize",
]



from era5_torch import (
    ContinuousChunker,
    SingleChunker,
    SingleRandomWrapChunker,
    Offset,
)
import pytest
import numpy as np
import xarray as xr


# fixtures
@pytest.fixture(scope="session")
def arr():
    data = np.array(
        [
            [
                [
                    [1, 2, 3, 4, 5, 6],
                    [7, 8, 9, 10, 11, 12],
                    [13, 14, 15, 16, 17, 18],
                    [19, 20, 21, 22, 23, 24],
                ]
            ]
        ]
    )
    ds = xr.Dataset(
        data_vars={
            "t": (
                ["level", "time", "latitude", "longitude"],
                data,
            )
        },
        coords={
            "level": [1000],
            "time": [0],
            "latitude": np.arange(4),
            "longitude": np.arange(6),
        },
    )
    return ds


@pytest.fixture(scope="session")
def arr3d():
    data = np.array(
        [
            [
                [
                    [1, 2, 3, 4, 5, 6],
                    [7, 8, 9, 10, 11, 12],
                    [13, 14, 15, 16, 17, 18],
                    [19, 20, 21, 22, 23, 24],
                ],
                [
                    [25, 26, 27, 28, 29, 30],
                    [31, 32, 33, 34, 35, 36],
                    [37, 38, 39, 40, 41, 42],
                    [43, 44, 45, 46, 47, 48],
                ],
                [
                    [49, 50, 51, 52, 53, 54],
                    [55, 56, 57, 58, 59, 60],
                    [61, 62, 63, 64, 65, 66],
                    [67, 68, 69, 70, 71, 72],
                ],
                [
                    [73, 74, 75, 76, 77, 78],
                    [79, 80, 81, 82, 83, 84],
                    [85, 86, 87, 88, 89, 90],
                    [91, 92, 93, 94, 95, 96],
                ],
            ]
        ]
    )
    # make an xarray dataset with the data and dimensions level time latitude longitude
    ds = xr.Dataset(
        data_vars={"t": (["level", "time", "latitude", "longitude"], data)},
        coords={
            "level": [1000],
            "time": np.arange(4),
            "latitude": np.arange(4),
            "longitude": np.arange(6),
        },
    )
    return ds


def test_chunks_2d(arr):
    chunk_shape = (2, 3)
    chunker = ContinuousChunker(chunk_shape)
    assert np.allclose(
        chunker.get_chunk_at(arr, 0)["t"].values, np.array([[1, 2, 3], [7, 8, 9]])
    )
    assert np.allclose(
        chunker.get_chunk_at(arr, 1)["t"].values,
        np.array([[13, 14, 15], [19, 20, 21]]),
    )
    assert np.allclose(
        chunker.get_chunk_at(arr, 2)["t"].values,
        np.array([[4, 5, 6], [10, 11, 12]]),
    )
    assert np.allclose(
        chunker.get_chunk_at(arr, 3)["t"].values,
        np.array([[16, 17, 18], [22, 23, 24]]),
    )
    assert chunker.get_num_chunks(arr) == 4


def test_chunks_3d(arr3d):
    chunk_shape = (2, 2, 3)
    chunker = ContinuousChunker(chunk_shape)
    assert np.allclose(
        chunker.get_chunk_at(arr3d, 0)["t"].values,
        np.array([[[1, 2, 3], [7, 8, 9]], [[25, 26, 27], [31, 32, 33]]]),
    )
    assert np.allclose(
        chunker.get_chunk_at(arr3d, 1)["t"].values,
        np.array([[[49, 50, 51], [55, 56, 57]], [[73, 74, 75], [79, 80, 81]]]),
    )
    assert np.allclose(
        chunker.get_chunk_at(arr3d, 2)["t"].values,
        np.array([[[13, 14, 15], [19, 20, 21]], [[37, 38, 39], [43, 44, 45]]]),
    )
    assert np.allclose(
        chunker.get_chunk_at(arr3d, 3)["t"].values,
        np.array([[[61, 62, 63], [67, 68, 69]], [[85, 86, 87], [91, 92, 93]]]),
    )
    assert np.allclose(
        chunker.get_chunk_at(arr3d, 4)["t"].values,
        np.array([[[4, 5, 6], [10, 11, 12]], [[28, 29, 30], [34, 35, 36]]]),
    )
    assert chunker.get_num_chunks(arr3d) == 8


def test_single_chunker_2d(arr):
    chunk_shape = (2, 3)
    chunker = SingleChunker(chunk_shape)
    assert np.allclose(
        chunker.get_chunk_at(arr, 0)["t"].values, np.array([[1, 2, 3], [7, 8, 9]])
    )
    assert chunker.get_num_chunks(arr) == 1


def test_single_chunker_3d(arr3d):
    chunk_shape = (2, 2, 3)
    chunker = SingleChunker(chunk_shape)
    assert np.allclose(
        chunker.get_chunk_at(arr3d, 0)["t"].values,
        np.array([[[1, 2, 3], [7, 8, 9]], [[25, 26, 27], [31, 32, 33]]]),
    )
    assert chunker.get_num_chunks(arr3d)


def test_single_chunker_throws(arr):
    chunker_shape = (2, 2, 3)
    chunker = SingleChunker(chunker_shape)
    with pytest.raises(ValueError):
        chunker.get_chunk_at(arr, 1)

    chunker_shape = (2, 2)
    chunker = SingleChunker(chunker_shape)
    with pytest.raises(ValueError):
        chunker.get_chunk_at(arr, 1)


def test_single_chunker_offset_2d(arr):
    chunk_shape = (2, 3)
    chunker = SingleChunker(chunk_shape, offset=[1, 1])
    assert np.allclose(
        chunker.get_chunk_at(arr, 0)["t"].values, np.array([[8, 9, 10], [14, 15, 16]])
    )
    assert chunker.get_num_chunks(arr) == 1


def test_single_chunker_offset_3d(arr3d):
    chunk_shape = (2, 2, 3)
    chunker = SingleChunker(chunk_shape, offset=[1, 1, 1])
    assert np.allclose(
        chunker.get_chunk_at(arr3d, 0)["t"].values,
        np.array([[[32, 33, 34], [38, 39, 40]], [[56, 57, 58], [62, 63, 64]]]),
    )
    assert chunker.get_num_chunks(arr3d) == 1


def test_random_chunker_2d(arr):
    chunk_shape = (2, 3)
    chunker = SingleRandomWrapChunker(chunk_shape, seed=0)
    assert chunker.get_num_chunks(arr) == 1
    arr_raw = arr["t"].values.squeeze()
    for _ in range(20):
        sample = chunker.get_chunk_at(arr, 0)["t"].values.squeeze()
        assert sample.shape == chunk_shape
        assert np.isin(sample, arr_raw).all()
        assert np.isin(sample, arr_raw).sum() == 6


def test_random_chunker_3d(arr3d):
    chunk_shape = (2, 2, 3)
    chunker = SingleRandomWrapChunker(chunk_shape, seed=0)
    arr_raw = arr3d["t"].values.squeeze()
    assert chunker.get_num_chunks(arr3d) == 1
    for _ in range(20):
        sample = chunker.get_chunk_at(arr3d, 0)["t"].values.squeeze()
        assert sample.shape == chunk_shape
        assert np.isin(sample, arr_raw).all()
        assert np.isin(sample, arr_raw).sum() == 12


def test_random_offset_continuous_chunker_2d(arr):
    chunk_shape = (2, 3)
    chunker = ContinuousChunker(chunk_shape, offset=Offset.RANDOM, seed=0)
    arr_raw = arr["t"].values.squeeze()
    assert chunker.get_num_chunks(arr) == 4
    mask = np.zeros(arr_raw.shape, dtype=bool)

    for i in range(4):
        sample = chunker.get_chunk_at(arr, i)["t"].values.squeeze()
        assert sample.shape == chunk_shape
        assert np.isin(sample, arr_raw).all()
        assert np.isin(sample, arr_raw).sum() == 6
        mask = np.logical_or(mask, np.isin(arr_raw, sample))

    assert mask.all()


def test_random_offset_continuous_chunker_3d(arr3d):
    chunk_shape = (2, 2, 3)
    chunker = ContinuousChunker(chunk_shape, offset=Offset.RANDOM, seed=0)
    arr_raw = arr3d["t"].values.squeeze()
    assert chunker.get_num_chunks(arr3d) == 8
    mask = np.zeros(arr_raw.shape, dtype=bool)

    for i in range(8):
        sample = chunker.get_chunk_at(arr3d, i)["t"].values.squeeze()
        assert sample.shape == chunk_shape
        assert np.isin(sample, arr_raw).all()
        assert np.isin(sample, arr_raw).sum() == 12
        mask = np.logical_or(mask, np.isin(arr_raw, sample))

    assert mask.all()



from __future__ import annotations
from typing import Optional
from torch.utils.data import Dataset
import torch
from os import PathLike
import numpy as np
import torchvision.transforms as T

from ._auxiliaries import (
    add_coords_information,
    add_land_sea_mask,
    Coordinate,
    CoordinateBundle,
)

from ._chunkers import Chunker
import xarray
from enum import Enum

import glob
from pathlib import Path


# taken from our large 2008 dataset
T_MIN = 194.8391418457031
T_MAX = 327.346343994141
T_MEAN = 264.752433
T_STD = 18.535514


class DatasetDimension(Enum):
    """
    Enum for the dimension of the dataset.
    """

    DIM2 = 0
    DIM3 = 1


def _sample_values_to_3d_tensor(sample: xarray.DataArray) -> torch.Tensor:
    """Gets a sample from the dataset.

    Returns:
        A sample from the dataset at the given index. The sample is a 3D tensor.
        If a 2D sample is requested, a first dimension is added to the tensor as (1, x, y).
    """
    tensor = torch.Tensor(sample.values)
    if len(tensor.shape) == 2:
        tensor = tensor.unsqueeze(0)
    assert len(tensor.shape) == 3
    return tensor


def normalize(arr: torch.Tensor | np.ndarray):
    """Normalizes the array to be in the range [0, 1]. Uses
    the precomputed values from the year 2008 that might not be appropriate for other datasets.
    """
    return (arr - T_MIN) / (T_MAX - T_MIN)


def unnormalize(arr: torch.Tensor | np.ndarray):
    """Undoes the normalization of this package. Uses the precomputed values from the year 2008 that might not be appropriate for other datasets."""
    return arr * (T_MAX - T_MIN) + T_MIN


class Era5(Dataset):
    """Class that wraps the Era5 dataset to be used with PyTorch.
    Expect the dataset to be in NetCDF format, and contain latitude, longitude, level, and time dimensions.

    The samples can be accessed via an array-like access: dataset[0] will return the first sample.
    Increasing the index will first increase chunks (if applicable), then the level (if not fixed), then the time, unless a specific level has been fixed.

    The samples are returned as 2D or 3D tensors, depending on the chunker.
    If no chunker is set, we only return 2D tensors (as returning a full 3D sample is likely to be too large).

    Args:
        path: Path to the dataset.
        data_variable: Name of the variable in the dataset that contains the data, for example 't' for temperature.
        normalize: Whether to normalize the data to be in the range [0, 1]. For the temperature, uses precomputed values from the year 2008 that might not be appropriate for other datasets.
            The normalization also affects the passed coordinates. If a tuple is passed, uses the first number as min, the second as max.
            Defaults to False.
        normalize_coords: Whether to normalize the coordinates to be in the range [0, 1]. Defaults to True.
        level: If set, only samples from this level will be returned. If not set, all levels will be returned. This integer is the index of the level, not the actual value.
        chunker: If set, the dataset will be chunked into smaller samples. If not set, the dataset will be returned as is.
        transform: A function that generates a torch tensor, which can be consumed by a neural network, from an xarray sample. In the default transform,
            the sample is achieved by taking the values of the xarray sample and converting it to a 3D torch tensor. All of the sample transform should return
            a 3D Tensor and add auxiliary information (if applicable) along the first dimension.
        coords: The coordinates to add to the sample. Coordinates are added in the order that they are passed to the first dimension of the returned sample.
    """

    def __init__(
        self,
        dataset: str | PathLike | xarray.Dataset,
        data_variable: str,
        normalize: bool | tuple[float, float] = False,
        normalize_coords: bool = True,
        level: Optional[int] = None,
        chunker: Optional[Chunker] = None,
        coords: list[Coordinate] | CoordinateBundle = [],
        transform=None,
    ) -> None:
        self._era5: _Era5Base
        self._data_variable = data_variable
        self._normalize_coords = normalize_coords
        if chunker is None:
            self._era5 = _Era5Base(dataset, data_variable, level)
        else:
            self._era5 = _Era5Chunked(dataset, data_variable, chunker, level)
        self._normalize = normalize
        self._coords = coords.value if isinstance(coords, CoordinateBundle) else coords
        self.transform = transform

    def __str__(self) -> str:
        return str(self._era5)

    def __repr__(self) -> str:
        return self.__str__()

    def _normalize_sample(self, sample: xarray.DataArray) -> xarray.DataArray:
        """
        Normalizes the sample so that it falls between 0 and 1.
        """
        if isinstance(self._normalize, tuple):
            return (sample - self._normalize[0]) / (
                self._normalize[1] - self._normalize[0]
            )
        if self._normalize:
            return (sample - T_MIN) / (T_MAX - T_MIN)
        else:
            return sample

    def _get_normalized_data_array(self, index: int) -> xarray.DataArray:
        return self._normalize_sample(self._era5[index][self._data_variable])

    def __getitem__(self, index: int) -> torch.Tensor:
        """Gets a sample from the dataset.

        Returns:
            A sample from the dataset at the given index.
        """
        sample = add_coords_information(
            self._get_normalized_data_array(index),
            self._coords,
            normalize=self._normalize_coords,
        )
        return self.transform(sample) if self.transform is not None else sample

    def __len__(self) -> int:
        """
        Returns:
            Number of samples in this dataset.
        """
        return self._era5.__len__()

    @property
    def bytes(self) -> int:
        return self._era5.bytes

    @property
    def numel(self) -> int:
        return self._era5.numel


class Era5Lsm(Era5):
    """
    Era5 with a land-sea mask added in the first dimension.
    """

    def __init__(
        self,
        dataset: str | PathLike | xarray.Dataset,
        lsm_dataset: str | PathLike,
        data_variable: str,
        normalize: bool = True,
        level: Optional[int] = None,
        chunker: Optional[Chunker] = None,
    ) -> None:
        super().__init__(dataset, data_variable, normalize, level, chunker)
        if isinstance(lsm_dataset, xarray.Dataset):
            self._lsm_dataset = lsm_dataset
        else:
            self._lsm_dataset = xarray.open_dataset(lsm_dataset)  # type: ignore

    def __getitem__(self, index: int) -> torch.Tensor:
        return add_land_sea_mask(
            self._get_normalized_data_array(index), self._lsm_dataset
        )


class _Era5Base:
    """Class for a simple, unchunked access to Era5. See the Era5 class for more information."""

    def __init__(
        self,
        dataset: str | PathLike | xarray.Dataset,
        data_variable: str,
        level: Optional[int] = None,
    ) -> None:
        if isinstance(dataset, xarray.Dataset):
            self._dataset = dataset
        else:
            self._dataset = xarray.open_dataset(dataset)  # type: ignore
        if level is not None:
            # keeps dimension
            self._dataset = self._dataset.sel(level=[level], method="nearest")
        if data_variable not in self._dataset.data_vars:
            raise ValueError(f"'{data_variable}' not found in {self._dataset}.")
        self._no_level = "level" not in self._dataset.coords
        if self._no_level:
            self._num_levels = 1
        else:
            self._num_levels = self._dataset["level"].size
        self._num_timesteps = self._dataset["time"].size
        self._num_latitude = self._dataset["latitude"].size
        self._num_longitude = self._dataset["longitude"].size
        self._variable = data_variable
        self._data_variable = data_variable
        self._level = level
        self._dimension = DatasetDimension.DIM2

    def check_index(self, index):
        if index >= len(self):
            raise ValueError(f"Index {index} is larger than dataset ({len(self)})")

    def __str__(self) -> str:
        return f"Era5 Base: {self._dataset}"

    def __repr__(self) -> str:
        return self.__str__()

    def _isel_level_safe(self, level: int, **kwargs) -> xarray.Dataset:
        """
        Helper function to handle datasets more cleanly that don't have a level dimension.
        """
        if self._no_level:
            return self._dataset.isel(**kwargs)
        else:
            return self._dataset.isel(level=level, **kwargs)

    def __getitem__(self, index) -> xarray.Dataset:
        self.check_index(index)
        level_index = index % self._num_levels
        time_index = index // self._num_levels

        sample = self._isel_level_safe(time=time_index, level=level_index)
        return sample

    def __len__(self) -> int:
        return self._num_timesteps * self._num_levels

    @property
    def bytes(self) -> int:
        if self._dataset[self._data_variable].dtype == np.float32:
            el_size = 4
        elif self._dataset[self._data_variable].dtype == np.float64:
            el_size = 8
        else:
            raise ValueError(
                f"Unknown dtype {self._dataset[self._data_variable].dtype}"
            )
        return el_size * self.numel

    @property
    def numel(self) -> int:
        return int(self._dataset[self._data_variable].size)


class _Era5Chunked(_Era5Base):
    """Class for chunked access to Era5. See the Era5 class for more information.
    Chunking means that each sample is split into several chunks, that together makeup the full sample.

    The type of chunker determines the dimension of the samples.
    """

    def __init__(
        self,
        dataset: str | PathLike | xarray.Dataset,
        data_variable: str,
        chunker: Chunker,
        level: Optional[int] = None,
    ) -> None:
        super().__init__(dataset, data_variable, level)
        self._chunker = chunker
        if self._chunker.dimension == 2:
            self._sample_shape = {
                "latitude": self._num_latitude,
                "longitude": self._num_longitude,
            }
        elif self._chunker.dimension == 3:
            self._sample_shape = {
                "time": self._num_timesteps,
                "latitude": self._num_latitude,
                "longitude": self._num_longitude,
            }
        else:
            raise ValueError(f"Invalid dimension {self._chunker.dimension}")
        self._chunks_per_sample = self._chunker.get_num_chunks(self._sample_shape)

    def __str__(self) -> str:
        return f"Era5 Chunked ({self._chunker}): {self._dataset}"

    def __repr__(self) -> str:
        return self.__str__()

    def __getitem__(self, index) -> xarray.Dataset:
        self.check_index(index)
        # eventually the samples are too large to simply draw them, we have to work with the coordinates directly.
        sample_index = index // self._chunks_per_sample
        chunk_index = index % self._chunks_per_sample
        if self._chunker.dimension == 2:
            # pre-select time and level for the chunker, so it can only chunk over longitude and latitude
            level_index = sample_index % self._num_levels
            time_index = sample_index // self._num_levels
            sample = self._isel_level_safe(time=time_index, level=level_index)
        elif self._chunker.dimension == 3:
            # in this case, chunker has to deal with the time index
            level_index = sample_index
            sample = self._isel_level_safe(level=level_index)
        else:
            raise ValueError(f"Invalid dimension {self._chunker.dimension}")
        chunk = self._chunker.get_chunk_at(sample, chunk_index)
        return chunk

    def __len__(self) -> int:
        if self._chunker.dimension == 2:
            return self._num_timesteps * self._num_levels * self._chunks_per_sample
        elif self._chunker.dimension == 3:
            return self._num_levels * self._chunks_per_sample
        else:
            raise ValueError(f"Invalid dimension {self._chunker.dimension}")



from __future__ import annotations
from typing import Optional, Hashable, Mapping
from enum import Enum
import xarray
from abc import ABC, abstractmethod
import numpy as np


class Offset(Enum):
    """Enum for the offset of the chunks."""

    NONE = 0
    RANDOM = 1


class Chunker(ABC):
    """Class that takes in image data and returns a chunked version of it.

    All of the chunks together make up one full image.

    Args:
        chunk_shape: The shape of the chunks. Might be 2D or 3D. Alternatively, a dictionary may be passed,
        which contains the keys latitude, longitude and optionally time, and as values the chunk shapes of the corresponding
        dimension
        padding: The padding type to use. Currently only NONE is implemented.
    """

    def __init__(self, chunk_shape: tuple | dict) -> None:
        self._chunk_shape: dict

        if isinstance(chunk_shape, tuple):
            if len(chunk_shape) == 2:
                self._chunk_shape = {
                    "latitude": chunk_shape[0],
                    "longitude": chunk_shape[1],
                }
            elif len(chunk_shape) == 3:
                self._chunk_shape = {
                    "time": chunk_shape[0],
                    "latitude": chunk_shape[1],
                    "longitude": chunk_shape[2],
                }
            else:
                raise ValueError(
                    f"Wrong length passed for chunk shape (must be 2 or 3): {chunk_shape}"
                )
        elif isinstance(chunk_shape, dict):
            self._chunk_shape = chunk_shape
        else:
            raise ValueError(f"Wrong type passed for chunk shape: {type(chunk_shape)}")

    def __str__(self) -> str:
        return f"Chunker {self._chunk_shape}"

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def dimension(self) -> int:
        """
        Returns:
            The dimension of the chunks, i.e. 2 or 3.
        """
        return len(self._chunk_shape.values())

    def _check_sample_validity(self, sample: xarray.Dataset, idx: int):
        if self.dimension == len(sample.coords.keys()):
            raise ValueError(
                f"Sample must be {self.dimension}D, but is {len(sample.coords.keys())}D"
            )
        if idx >= self.get_num_chunks(sample.shape):
            raise ValueError(
                f"Index {idx} out of bounds, max index is {self.get_num_chunks(sample.shape) - 1}"
            )

    @abstractmethod
    def get_chunk_at(self, sample: xarray.Dataset, idx: int) -> xarray.Dataset:
        """Returns the chunk at the given index for the sample.

        Args:
            sample: The sample to get the chunk from.
            idx: The index of the chunk to get.
        """
        raise NotImplementedError()

    @abstractmethod
    # convoluted type hinting for sample shape is required, since xarray.Dataset.coords.dims is a Mapping[Hashable, int], which
    # is invariant in its first type
    def get_num_chunks(
        self, sample_shape: dict[str, int] | xarray.Dataset | Mapping[Hashable, int]
    ) -> int:
        """Returns the number of possible chunks in a sample of the given shape.

        Args:
            sample_shape: The shape of the sample.
        """
        raise NotImplementedError()


class ContinuousChunker(Chunker):
    """Chunker that chunks the image into a continuous grid of chunks. The chunks don't overlap,
    and they make up the full image together.
    """

    def __init__(
        self,
        chunk_shape: tuple | dict,
        offset: Offset = Offset.NONE,
        seed: Optional[int] = None,
    ) -> None:
        super().__init__(chunk_shape)
        self._offset = None
        self._offset_mode = offset
        self._seed = seed

    def _determine_offset(self, sample: xarray.Dataset):
        self._offset = {}
        if self._offset_mode == Offset.RANDOM:
            self.rng = np.random.default_rng(self._seed)
            # we had to defer the off set generation to here, because we don't know the sample shape yet
            for coord_name, coord_size in sample.coords.dims.items():
                self._offset[coord_name] = np.random.randint(coord_size)
        elif self._offset_mode == Offset.NONE:
            for coord_name, coord_size in sample.coords.dims.items():
                self._offset[coord_name] = 0
        else:
            raise ValueError(f"Unknown offset mode {self._offset_mode}")

    def get_chunk_at(self, sample: xarray.Dataset, idx: int) -> xarray.Dataset:
        # this has to be done in a lazy fashion, because we don't know the sample shape yet
        if self._offset is None:
            self._determine_offset(sample)
        assert self._offset is not None

        for coord_name, coord_value in self._chunk_shape.items():
            coordinate = sample[coord_name]
            dim_size = coordinate.size
            offset = self._offset[coord_name]

            num_entries = dim_size // coord_value
            if num_entries == 0:  # empty coordinate
                continue
            idx, current = divmod(idx, num_entries)

            start = (current * coord_value + offset) % dim_size
            end = ((current + 1) * coord_value + offset) % dim_size

            start_val = coordinate[start]
            end_val = coordinate[end]

            if end == 0:  # else the wrap around is incorrect
                end = dim_size

            bigger_equal_start = coordinate >= start_val
            smaller_than_end = coordinate < end_val

            if start < end:
                sample = sample.isel({coord_name: slice(start, end)})
            else:  # wrap-around case
                sample = sample.isel(
                    {coord_name: bigger_equal_start | smaller_than_end}
                )
        return sample

    def get_num_chunks(
        self, sample_shape: dict[str, int] | xarray.Dataset | Mapping[Hashable, int]
    ) -> int:
        if isinstance(sample_shape, xarray.Dataset):
            sample_shape = sample_shape.coords.dims
        num_chunks = 1
        for coord_name, coord_size in sample_shape.items():
            if coord_name not in self._chunk_shape:
                continue
            assert isinstance(
                coord_name, str
            )  # it should normally be a str anyway, unless you do weird things
            num_chunks *= coord_size // self._chunk_shape[coord_name]
        return num_chunks


class SingleRandomWrapChunker(ContinuousChunker):
    """Chunker that returns a random chunker per image. We assume that the data lives on a spherical
    surface and that its coordinates thus wrap around (e.g. coords[0] is adjacent to coords[-1]).
    The returned chunk may wrap around coordinate borders.
    """

    def __init__(self, chunk_shape: tuple, seed: Optional[int] = None) -> None:
        super().__init__(chunk_shape, offset=Offset.RANDOM)

    def get_chunk_at(self, sample: xarray.Dataset, idx: int) -> xarray.Dataset:
        assert idx == 0
        return super().get_chunk_at(sample, idx)

    def get_num_chunks(self, _) -> int:
        return 1


class SingleChunker(Chunker):
    """Simple Chunker that returns only a single chunk per image. See the base class for more information.

    Args:
        offset: The offset of the chunk. If None, the chunk will start at 0 in all dimensions. If specified,
        the offset must have the same length as the chunk shape.
    """

    def __init__(
        self,
        chunk_shape: tuple[int, ...],
        offset: Optional[list[int] | dict] = None,
    ) -> None:
        super().__init__(chunk_shape)
        if offset is None:
            self._offset = {k: 0 for k, _ in self._chunk_shape.items()}
        elif isinstance(offset, list):
            if len(offset) == 2:
                self._offset = {"latitude": offset[0], "longitude": offset[1]}
            elif len(offset) == 3:
                self._offset = {
                    "time": offset[0],
                    "latitude": offset[1],
                    "longitude": offset[2],
                }
        else:
            self._offset = offset

    def get_chunk_at(self, sample: xarray.Dataset, idx: int) -> xarray.Dataset:
        if idx != 0:
            raise ValueError(f"Index {idx} out of bounds, max index is 0")
        for coord_name, coord_value in self._chunk_shape.items():
            coord_offset = self._offset[coord_name]
            sample = sample.isel(
                {coord_name: slice(coord_offset, coord_value + coord_offset)}
            )
        return sample

    def get_num_chunks(self, _) -> int:
        return 1



from abc import ABC, abstractmethod
import torch
from enum import Enum


class PatchDimension(Enum):
    """
    Enum for specifying the dimension of a patch.
    """

    DIM2 = 1
    DIM3 = 2


class Patcher(ABC):
    """
    Class that extracts patches from an image.
    """

    def __init__(self, patch_shape: tuple) -> None:
        if len(patch_shape) == 2:
            self._patch_dimension = PatchDimension.DIM2
        elif len(patch_shape) == 3:
            self._patch_dimension = PatchDimension.DIM3
        else:
            raise ValueError("Patch must be 2D or 3D")
        self._patch_shape = patch_shape

    def _validate_sample_shape(self, sample: torch.Tensor) -> None:
        if len(sample.shape) != len(self._patch_shape):
            raise ValueError(f"Sample must be {len(self._patch_shape)}D")
        for i in range(len(sample.shape)):
            if sample.shape[i] < self._patch_shape[i]:
                raise ValueError(
                    f"Sample must be at least {self._patch_shape[i]} in dimension {i}"
                )

    @abstractmethod
    def __call__(self, sample: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


class BeginningPatcher(Patcher):
    """
    Returns a patch from the start of the array.
    """

    def __init__(self, patch_shape: tuple) -> None:
        super().__init__(patch_shape)

    def __call__(self, sample: torch.Tensor) -> torch.Tensor:
        super()._validate_sample_shape(sample)
        if self._patch_dimension == PatchDimension.DIM2:
            patched_sample = sample[0 : self._patch_shape[0], 0 : self._patch_shape[1]]
        elif self._patch_dimension == PatchDimension.DIM3:
            patched_sample = sample[
                0 : self._patch_shape[0],
                0 : self._patch_shape[1],
                0 : self._patch_shape[2],
            ]
        else:
            raise ValueError("Patch must be 2D or 3D")
        assert patched_sample.shape == self._patch_shape
        return patched_sample



from era5_torch import (
    Era5,
    ContinuousChunker,
    Coordinate,
    CoordinateBundle,
    normalize,
    unnormalize,
)
import numpy as np
import torch
import xarray as xr
import pytest

temperature_path = "tests/era5_temperature_small.nc"
lsm_path = "tests/era5_lsm_small.nc"


def test_normalize_unnormalize_cancels():
    for i in range(100):
        xs = torch.rand(2, 3) * 100
        xs_prime = normalize(unnormalize(xs))
        xs_prime_reverse = unnormalize(normalize(xs))
        assert np.allclose(xs, xs_prime, atol=1e-4)
        assert np.allclose(xs_prime_reverse, xs_prime, atol=1e-4)
        assert np.allclose(xs_prime_reverse, xs, atol=1e-4)


def test_era5_retrieval_correct_values():
    ds = xr.open_dataset(temperature_path)
    era5 = Era5(temperature_path, "t", normalize=False)

    assert era5[0].squeeze().shape == (ds.latitude.size, ds.longitude.size)
    assert torch.allclose(
        era5[0].squeeze(), torch.Tensor(ds.t.isel(time=0, level=0).values)
    )
    assert torch.allclose(
        era5[1].squeeze(), torch.Tensor(ds.t.isel(time=0, level=1).values)
    )
    assert torch.allclose(
        era5[3].squeeze(), torch.Tensor(ds.t.isel(time=1, level=0).values)
    )


def test_era5_retrieval_correct_values_fixed_level():
    ds = xr.open_dataset(temperature_path)
    era5 = Era5(temperature_path, "t", normalize=False, level=1000)

    assert era5[0].squeeze().shape == (ds.latitude.size, ds.longitude.size)
    assert torch.allclose(
        era5[0].squeeze(), torch.Tensor(ds.t.isel(time=0, level=0).values)
    )
    assert torch.allclose(
        era5[1].squeeze(), torch.Tensor(ds.t.isel(time=1, level=0).values)
    )
    assert torch.allclose(
        era5[2].squeeze(), torch.Tensor(ds.t.isel(time=2, level=0).values)
    )


def test_era5_chunked_retrieval_correct_values_2d():
    ds = xr.open_dataset(temperature_path)
    era5 = Era5(
        temperature_path, "t", normalize=False, chunker=ContinuousChunker((7, 7))
    )

    assert torch.allclose(
        era5[0],
        torch.Tensor(
            ds.t.isel(
                time=0, level=0, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[1],
        torch.Tensor(
            ds.t.isel(
                time=0, level=0, latitude=slice(7, 14), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[3],
        torch.Tensor(
            ds.t.isel(
                time=0, level=0, latitude=slice(0, 7), longitude=slice(7, 14)
            ).values
        ),
    )
    assert torch.allclose(
        era5[9],
        torch.Tensor(
            ds.t.isel(
                time=0, level=1, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )


def test_era5_chunked_retrieval_correct_values_fixed_level():
    ds = xr.open_dataset(temperature_path)
    era5 = Era5(
        temperature_path,
        "t",
        normalize=False,
        chunker=ContinuousChunker((7, 7)),
        level=1,
    )

    assert torch.allclose(
        era5[0],
        torch.Tensor(
            ds.t.isel(
                time=0, level=2, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[1],
        torch.Tensor(
            ds.t.isel(
                time=0, level=2, latitude=slice(7, 14), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[3],
        torch.Tensor(
            ds.t.isel(
                time=0, level=2, latitude=slice(0, 7), longitude=slice(7, 14)
            ).values
        ),
    )
    assert torch.allclose(
        era5[9],
        torch.Tensor(
            ds.t.isel(
                time=1, level=2, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )


def test_era5_chunked_retrieval_correct_values_3d():
    ds = xr.open_dataset(temperature_path)
    era5 = Era5(
        temperature_path, "t", normalize=False, chunker=ContinuousChunker((2, 7, 7))
    )

    assert torch.allclose(
        era5[0],
        torch.Tensor(
            ds.t.isel(
                time=slice(0, 2), level=0, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[1],
        torch.Tensor(
            ds.t.isel(
                time=slice(2, 4), level=0, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[2],
        torch.Tensor(
            ds.t.isel(
                time=slice(4, 6), level=0, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[4],
        torch.Tensor(
            ds.t.isel(
                time=slice(0, 2), level=0, latitude=slice(7, 14), longitude=slice(0, 7)
            ).values
        ),
    )


def test_era5_chunked_retrieval_correct_values_fixed_level_3d():
    ds = xr.open_dataset(temperature_path)
    era5 = Era5(
        temperature_path,
        "t",
        level=1,
        normalize=False,
        chunker=ContinuousChunker((2, 7, 7)),
    )

    assert torch.allclose(
        era5[0],
        torch.Tensor(
            ds.t.isel(
                time=slice(0, 2), level=2, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[1],
        torch.Tensor(
            ds.t.isel(
                time=slice(2, 4), level=2, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[2],
        torch.Tensor(
            ds.t.isel(
                time=slice(4, 6), level=2, latitude=slice(0, 7), longitude=slice(0, 7)
            ).values
        ),
    )
    assert torch.allclose(
        era5[4],
        torch.Tensor(
            ds.t.isel(
                time=slice(0, 2), level=2, latitude=slice(7, 14), longitude=slice(0, 7)
            ).values
        ),
    )


def test_load_seamap():
    ds = xr.open_dataset(lsm_path)
    era5 = Era5(lsm_path, "lsm", normalize=False)

    assert era5[0].squeeze().shape == (ds.latitude.size, ds.longitude.size)
    assert torch.allclose(era5[0].squeeze(), torch.Tensor(ds.lsm.isel(time=0).values))
    assert torch.allclose(era5[1].squeeze(), torch.Tensor(ds.lsm.isel(time=1).values))
    assert torch.allclose(era5[3].squeeze(), torch.Tensor(ds.lsm.isel(time=2).values))


# fixtures
@pytest.fixture(scope="session")
def mock_dataset(tmp_path_factory):
    path = tmp_path_factory.mktemp("data") / "mock_dataset.nc"
    a = torch.Tensor([[1, 2, 3], [4, 5, 6]]).unsqueeze(0).unsqueeze(0)
    latitude = np.array([10, 20])
    longitude = np.array([30, 40, 50])
    time = np.array([-1])
    level = np.array([-10])
    ds = xr.Dataset(
        data_vars={"t": (("level", "time", "latitude", "longitude"), a)},
        coords={
            "level": level,
            "longitude": longitude,
            "latitude": latitude,
            "time": time,
        },
    )
    ds.to_netcdf(str(path))
    return path


# fixtures
@pytest.fixture(scope="session")
def mock_dataset_euclidean(tmp_path_factory):
    path = tmp_path_factory.mktemp("data") / "mock_dataset_euclidean.nc"
    a = torch.Tensor([[1, 2, 3], [4, 5, 6]]).unsqueeze(0).unsqueeze(0)
    latitude = np.array([90, 0])
    longitude = np.array([0, 90, 180])
    time = np.array([-1])
    level = np.array([-10])
    ds = xr.Dataset(
        data_vars={"t": (("level", "time", "latitude", "longitude"), a)},
        coords={
            "level": level,
            "longitude": longitude,
            "latitude": latitude,
            "time": time,
        },
    )
    ds.to_netcdf(str(path))
    return path


# fixtures
@pytest.fixture(scope="session")
def mock_dataset_lengths(tmp_path_factory):
    path = tmp_path_factory.mktemp("data") / "mock_dataset_euclidean.nc"
    a = np.random.random((1, 21, 60, 90))
    latitude = np.arange(60)
    longitude = np.arange(90)
    time = np.arange(21)
    level = np.array([-10])
    ds = xr.Dataset(
        data_vars={"t": (("level", "time", "latitude", "longitude"), a)},
        coords={
            "level": level,
            "longitude": longitude,
            "latitude": latitude,
            "time": time,
        },
    )
    ds.to_netcdf(str(path))
    return path


def test_dataset_coords_sphere(mock_dataset):
    ds = xr.open_dataset(mock_dataset)
    era5 = Era5(
        ds, "t", normalize=False, coords=CoordinateBundle.SPHERE, normalize_coords=False
    )

    assert era5[0].shape == (5, 2, 3)
    assert torch.allclose(
        era5[0],
        torch.Tensor(
            [
                [[1, 2, 3], [4, 5, 6]],
                [[-10, -10, -10], [-10, -10, -10]],
                [[-1, -1, -1], [-1, -1, -1]],
                [[10, 10, 10], [20, 20, 20]],
                [[30, 40, 50], [30, 40, 50]],
            ]
        ),
    )


def test_dataset_coords_euclidean(mock_dataset_euclidean):
    ds = xr.open_dataset(mock_dataset_euclidean)
    era5 = Era5(
        ds,
        "t",
        normalize=False,
        coords=CoordinateBundle.EUCLIDEAN,
        normalize_coords=False,
    )

    # just saving some space
    t0 = torch.Tensor([0])
    t1 = torch.Tensor([1])
    x00, y00, z00 = t0, t0, t1
    x01, y01, z01 = t0, t0, t1
    x02, y02, z02 = t0, t0, t1
    x10, y10, z10 = t1, t0, t0
    x11, y11, z11 = t0, t1, t0
    x12, y12, z12 = -t1, t0, t0
    x = [[x00, x01, x02], [x10, x11, x12]]
    y = [[y00, y01, y02], [y10, y11, y12]]
    z = [[z00, z01, z02], [z10, z11, z12]]
    assert torch.allclose(
        era5[0],
        torch.Tensor(
            [
                [[1, 2, 3], [4, 5, 6]],
                [[-10, -10, -10], [-10, -10, -10]],
                [[-1, -1, -1], [-1, -1, -1]],
                x,
                y,
                z,
            ]
        ),
        atol=1e-6,
    )


def test_dataset_correct_length_unchunked(mock_dataset_lengths):
    ds = xr.open_dataset(mock_dataset_lengths)
    era5 = Era5(ds, "t", normalize=False)

    assert len(era5) == 21


def test_dataset_correct_length_chunked_2d(mock_dataset_lengths):
    ds = xr.open_dataset(mock_dataset_lengths)
    era5 = Era5(ds, "t", normalize=False, chunker=ContinuousChunker((2, 2)))
    assert len(era5) == 21 * 30 * 45

    era5 = Era5(ds, "t", normalize=False, chunker=ContinuousChunker((10, 10)))
    assert len(era5) == 21 * 6 * 9


def test_dataset_correct_length_chunked_3d(mock_dataset_lengths):
    ds = xr.open_dataset(mock_dataset_lengths)
    era5 = Era5(ds, "t", normalize=False, chunker=ContinuousChunker((3, 2, 2)))
    assert len(era5) == 7 * 30 * 45

    era5 = Era5(ds, "t", normalize=False, chunker=ContinuousChunker((7, 10, 10)))
    assert len(era5) == 3 * 6 * 9


def test_dataset_coordinates_specifying_sphere(mock_dataset):
    ds = xr.open_dataset(mock_dataset)
    era5 = Era5(
        ds,
        "t",
        normalize=False,
        coords=[Coordinate.LATITUDE, Coordinate.LONGITUDE],
        normalize_coords=False,
    )

    assert era5[0].shape == (3, 2, 3)
    assert torch.allclose(
        era5[0],
        torch.Tensor(
            [
                [[1, 2, 3], [4, 5, 6]],
                [[10, 10, 10], [20, 20, 20]],
                [[30, 40, 50], [30, 40, 50]],
            ]
        ),
    )

    era5 = Era5(
        ds,
        "t",
        normalize=False,
        coords=[Coordinate.TIME, Coordinate.LEVEL],
        normalize_coords=False,
    )

    assert era5[0].shape == (3, 2, 3)
    assert torch.allclose(
        era5[0],
        torch.Tensor(
            [
                [[1, 2, 3], [4, 5, 6]],
                [[-1, -1, -1], [-1, -1, -1]],
                [[-10, -10, -10], [-10, -10, -10]],
            ]
        ),
    )


def test_dataset_coordinates_specifying(mock_dataset_euclidean):
    ds = xr.open_dataset(mock_dataset_euclidean)
    era5 = Era5(
        ds,
        "t",
        coords=[Coordinate.Y, Coordinate.X],
        normalize=False,
        normalize_coords=False,
    )
    t0 = torch.Tensor([0])
    t1 = torch.Tensor([1])
    x00, y00 = t0, t0
    x01, y01 = t0, t0
    x02, y02 = t0, t0
    x10, y10 = t1, t0
    x11, y11 = t0, t1
    x12, y12 = -t1, t0
    x = [[x00, x01, x02], [x10, x11, x12]]
    y = [[y00, y01, y02], [y10, y11, y12]]
    assert torch.allclose(
        era5[0],
        torch.Tensor([[[1, 2, 3], [4, 5, 6]], y, x]),
        atol=1e-6,
    )



from typing import Optional
import torch
import xarray
from enum import Enum
import math
import numpy as np


class Coordinate(Enum):
    LATITUDE = "latitude"
    LONGITUDE = "longitude"
    LEVEL = "level"
    TIME = "time"
    X = "x"
    Y = "y"
    Z = "z"
    NULL = "null"


class CoordinateBundle(Enum):
    SPHERE = [
        Coordinate.LEVEL,
        Coordinate.TIME,
        Coordinate.LATITUDE,
        Coordinate.LONGITUDE,
    ]
    EUCLIDEAN = [
        Coordinate.LEVEL,
        Coordinate.TIME,
        Coordinate.X,
        Coordinate.Y,
        Coordinate.Z,
    ]


class CoordinateBundleTyper(Enum):
    """Typer enum for coordinate bundles, as the list value of the original enum raises an error."""

    SPHERE = "sphere"
    EUCLIDEAN = "euclidean"

    def to_enum(self):
        return CoordinateBundle[self.value.upper()]


def remove_auxiliary(x: torch.Tensor) -> torch.Tensor:
    """Default implementation to remove the auxiliary information from the tensor. Assumes that the auxiliary information is
    stored in additional channels along the first dimension.

    Args:
        x: The tensor with added auxiliary information. Auxiliary information must have been added from the same class.

    Returns:
        The tensor without auxiliary information.
    """
    return x[0]


def add_coords_information(
    sample: xarray.DataArray,
    coords: list[Coordinate],
    normalize: bool = True,
) -> torch.Tensor:
    """Transforms sample to a tensor with auxiliary information added along the first dimension.
    The sample must have only a single level and timestamp.
    The input tensor must have shape (1,x,y) or (x,y), the output shape will be (n,x,y), where n = len(coords).
    The spatial information will be added as a meshgrid in the first dimension, time and level time will be added flat to all inputs.
    For example for coordinates [level, time, latitude, longitude] the output a will be (4,x,y), where the first dimension contains the level, time, latitude and longitude information.
    a[1,x,y] will give the level of the sample, a[2, x, y] will give the time of the sample, a[3, x, y] will give the latitude at coordinates (x,y) and a[4, x, y] will give the
    the longitude at coordinates (x,y).

    In the future, we might implement this function for samples that run over multiple time stamps.

    Args:
        sample: The sample to add auxiliary information to.
        normalize: Whether to normalize the coordinates to be in the range [0, 1]. This should always be on, except if you have a very valid reason to do so (i.e. debugging),
        as the size of some coordinates (particularly time) is very large and can thus disrupt the gradients when training the model.
        coords: The coordinates to add to the sample. The order of the coordinates is important, as this represents the order of the auxiliary information in the output tensor.

    Returns:
        A tensor representation of the sample with the auxiliary information added in the first dimension of shape (n,width,height). The coordinates will be added in the order that coords specifies.
    """
    a = torch.Tensor(sample.values)
    if len(a.shape) == 2:
        a = a.unsqueeze(0)
    if len(coords) == 0:
        # shortcut for no added coordinates
        return a

    latitude = torch.Tensor(sample.latitude.values)
    longitude = torch.Tensor(sample.longitude.values)
    time = sample.time.values
    level = sample.level.values

    if time.size != 1:
        raise ValueError(
            f"Only a single time stamp is supported at the moment, input has {time.size}"
        )
    if level.size != 1:
        raise ValueError(
            f"The sample must have only a single level, input has {time.size}"
        )

    if normalize:
        level_value = (level / 1000).item()
        time_value = (
            time.item() / np.datetime64("2009-01-01T00:00:00.000000000", "ns").item()
        )
    else:
        level_value = level.item()
        time_value = time.item()
    latitudes_view = torch.broadcast_to(torch.Tensor(latitude).view(1, -1, 1), a.shape)
    longitudes_view = torch.broadcast_to(
        torch.Tensor(longitude).view(1, 1, -1), a.shape
    )
    times_view = torch.full_like(a, time_value)
    level_view = torch.full_like(a, level_value)

    possible_coords = {"level": level_view, "time": times_view}
    # making lat-lon grid
    if normalize:
        possible_coords["latitude"] = latitudes_view / 180 + 1 / 2
        possible_coords["longitude"] = longitudes_view / 360 + 1 / 2
    else:
        possible_coords["latitude"] = latitudes_view
        possible_coords["longitude"] = longitudes_view

    # debug coordinates
    possible_coords["null"] = torch.full_like(a, 0)

    # making x-y-z grid
    x, y, z = transform_polar_to_euclidean(latitudes_view, longitudes_view)
    if normalize:
        x = x / 2 + 1 / 2
        y = y / 2 + 1 / 2
        z = z / 2 + 1 / 2
    possible_coords["x"] = x
    possible_coords["y"] = y
    possible_coords["z"] = z

    auxiliary_information = []
    for coord in coords:
        # coord.value just gives us the name of the coord, as coord is an enum
        coord_value = possible_coords.get(coord.value)
        if coord_value is None:
            # this should never happen to a user, as we use an enum
            raise ValueError(f"INTERNAL ERROR: Coordinate {coord} is not supported.")
        else:
            auxiliary_information.append(possible_coords[coord.value])

    return torch.cat([a] + auxiliary_information, dim=0)


def add_land_sea_mask(x: xarray.DataArray, lsm_dataset: xarray.Dataset) -> torch.Tensor:
    """
    Class for a callable sample transform, that adds a land-sea mask to the sample in the first dimension.

    Args:
        lsm_path: The path to the land-sea mask file. The file must be in netcdf format and must contain a variable
        named lsm, that contains the land-sea mask. The mask must contain the exact same coordinates as the origin data.
    """
    if "time" in x.coords:
        lsm = lsm_dataset.sel(
            time=x.time,
            latitude=x.latitude,
            longitude=x.longitude,
            method="nearest",
        )
    else:
        lsm = lsm_dataset.sel(
            latitude=x.latitude, longitude=x.longitude, method="nearest"
        )
    lsm_tensor = torch.Tensor(lsm["lsm"].values).unsqueeze(0)
    x_tensor = torch.Tensor(x.values)
    if len(x_tensor.shape) < len(lsm_tensor.shape):
        x_tensor = x_tensor.unsqueeze(0)
    if len(x_tensor.shape) != len(lsm_tensor.shape):
        raise ValueError("The shape of the sample and the land-sea mask do not match.")
    return torch.cat([x_tensor, lsm_tensor], 0)


def transform_polar_to_euclidean(
    lat: torch.Tensor, lng: torch.Tensor
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Transforms polar coordinates (latitude, longitude) to euclidean coordinates.
    Both sets of coordinate arrays must have the exact same shape. Each pair of coordinates
    (lat[i],long[i]) produces one triple (x[i],y[i],z[i])

    We assume that the spherical coordinates follow standard latitude/longitude conventions, e.g.
    the latitude is 0 on the equator and 90 on the northpole.

    Args:
        lat: Tensor containing the latitude in degrees, values range from [-90,90]
        longitude: Tensor containing the longitude in degrees, values range from [-180,180]
    """
    # the radius of earth is set to 1 wlog
    if lat.shape != lng.shape:
        raise ValueError(f"Shapes don't match: {lat.shape} != {lng.shape}")
    r = 1
    # conversion to radians
    # transform latitude: physics convention is that theta is the inclination angle from the north pole
    # towards the equator
    lat_r = (-lat + 90) / 360 * 2 * math.pi
    lng_r = lng / 360 * 2 * math.pi
    # ---
    x = r * torch.cos(lng_r) * torch.sin(lat_r)
    y = r * torch.sin(lng_r) * torch.sin(lat_r)
    z = r * torch.cos(lat_r)
    return x, y, z
